#include <cstdio>
#include "snes9x.h"
#include "libspc.h"
#include "soundux.h" // s9xexit uses
#include "apu.h"     // s9xexit uses

#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <string>

#include <termio.h>

using namespace std;

void S9xMessage (int type, int number, const char *message)
{
    type=type, number=number;
    fprintf (stderr, "%s\n", message);
}
bool8 S9xOpenSoundDevice (int mode, bool8 stereo, int buffer_size)
{
    mode=mode, stereo=stereo, buffer_size=buffer_size;
    return 0;
}

int S9xAPUOPrint (char *buffer, uint16 Address)
{
    buffer=buffer;
    Address=Address;
    return 0;
}
int S9xTraceAPU ()
{
    return 0;
}
void S9xTraceSoundDSP (const char *s, int i1 = 0, int i2 = 0, int i3 = 0,
                          int i4 = 0, int i5 = 0, int i6 = 0, int i7 = 0)
{
    fprintf(stderr, s, i1, i2, i3, i4, i5, i6, i7);
}


static struct termio back;

void S9xExit ()
{
    ioctl(0, TCSETA, &back);
    S9xSetSoundMute (TRUE);
    S9xDeinitAPU ();
    exit(0);
}

#include "argh.hh"

#define VERSIONSTR \
    "spcplay - SPC player based on snes9x\n" \
    "snes9x copyright http://www.snes9x.com/,\n" \
    "player copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "Thanks for the xmms spc plugin authors.\n"

static string filee;
static SPC_Config cfg = { 44100, 16, 2, 1, 1 };
static string snd_dev = "/dev/dsp";
static int surround = 0;

class Handle : public arghandler
{
    bool Help;
private:
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_V(const string &)
    {
        printf(VERSIONSTR);
        exit(0);
    }
    string opt_2(const string &s) { cfg.channels = 2;                   return s; }
    string opt_m(const string &s) { cfg.channels = 1;                   return s; }
    string opt_b(const string &s) { cfg.resolution = 8;                 return s; }
    string opt_6(const string &s) { cfg.resolution = 16;                return s; }
    string opt_n(const string &s) { cfg.is_interpolation = 0;           return s; }
    string opt_i(const string &s) { cfg.is_interpolation = 1;           return s; }
    string opt_e(const string &s) { cfg.is_echo = 1;                    return s; }
    string opt_c(const string &s) { cfg.is_echo = 0;                    return s; }
    string opt_a(const string &s) { Settings.AltSampleDecode = TRUE;    return s; }
    string opt_u(const string &s) { surround = 1;                       return s; }
    string opt_r(const string &s)
    {
        const char *q = s.c_str();
        const char *p = q;
        cfg.sampling_rate = strtol(p, (char **)&p, 10);
        if(cfg.sampling_rate < 18
        || cfg.sampling_rate > 256000)
        {
            argerror(s);
        }
        return s.substr(p-q);
    }
    string opt_o(const string &s)
    {
        snd_dev = s;
        return "";
    }
public:
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        Help(false)
    {
        add("-2", "--stereo", "Specifies stereo sound. Redundant. Default.", &Handle::opt_2);
        add("-m", "--mono",   "Specifies mono sound.",                       &Handle::opt_m);
        add("-b", "--byte",   "8-bit mode. Normally 16-bit mode.",           &Handle::opt_b);
        add("-6", "--word",   "16-bit mode. Redundant.",                     &Handle::opt_6);
        add("-r", "--rate",   "Specifies playing rate. 44410 is default.",   &Handle::opt_r);
        add("-i", "--ipolate","Enables interpolating. Default.",             &Handle::opt_i);
        add("-n", "--npolate","Disables interpolating. (Toggle-key: 9/i)",   &Handle::opt_n);
        add("-e", "--echo",   "Enables echo. Default.",                      &Handle::opt_e);
        add("-c", "--noecho", "Disables echo. (Toggle-key: e)",              &Handle::opt_c);
        add("-o", "--output", "Specify output device (default: /dev/dsp).",  &Handle::opt_o);
        add("-h", "--help",   "Help",                                        &Handle::opt_h);
        add("-V", "--version","Version information",                         &Handle::opt_V);
        add("-a", "--altsampledecode",
                               "Better sample decoding (Toggle-key: a)",         &Handle::opt_a);
        add("-u", "--surround","Flip one channel (stereo only) (Toggle-key: u)", &Handle::opt_u);
    }
    virtual void defarg(const string &arg)
    {
    	if(arg.size())filee = arg;
    }
    void Usage()
    {
    	fprintf(stderr, VERSIONSTR);
        fprintf(stderr,
            "Usage: %s [<options> ...] <file>\n"
            "\n"
            "Emulates and plays SPC files (snes soundtracks).\n"
            "\n"
            "Options:\n", a0.c_str());
        
        listoptions();
        
        exit(0);
    }
    virtual void parse()
    {
        arghandler::parse();
        if(Help || !filee.size())Usage();
    }
};

static double mastervolume = 1;

static void State(void)
{
     fprintf(stderr, "\rRate %-4d   Volume %-4.0f ", cur_rate, (mastervolume*100+0.5));
     for(unsigned a=0; a<8; ++a)
         fputc((so.sound_switch&(1<<a)) ? (a+'1') : '_', stderr);
     fflush(stderr);
}

static void KeyInit()
{
    struct termio term = back;
    
    // Disable linebuffer and echoing
    term.c_lflag &= ~(ICANON|ECHO);
    // Disable signals generated by ^C, ^Z and ^] (int,stop,quit)
    //term.c_lflag &= ~(ISIG);
    
    // Disable special handling of ^S and ^Q
    //term.c_iflag &= ~(IXON);
    
    term.c_cc[VMIN] = 0; // 0=no block, 1=do block
    ioctl(0, TCSETA, &term);
}

static void Resume(int)
{
	KeyInit();
}

int main(int argc, const char *const *argv)
{
    Handle parameters("", argc, argv);
    parameters.parse();
    
    ioctl(0, TCGETA, &back);
    signal(SIGCONT, Resume);
    KeyInit();

    unsigned bufsize = SPC_init(&cfg);
    unsigned char *Buf = (unsigned char *)malloc(bufsize);
    
#ifdef DEBUGGER
    Settings.TraceSoundDSP = 1;
#endif

#if 0
    for(unsigned fni=0; fni<files.size(); ++fni)
#else
	for(unsigned i=0; i<1; ++i)
#endif
    {
#if 0
        const char *fn = files[fni].c_str();
#else
		const char *fn = filee.c_str();
#endif
        if(!SPC_load(fn))
        {
            string s = fn;
            s += " could not be opened";
            S9xMessage(0,0, s.c_str());
            break;
        }
        
        int fd;
        
        if(snd_dev=="-" || !snd_dev.size())
            fd = STDOUT_FILENO;
        else
        {
            fd = open(snd_dev.c_str(), O_WRONLY);
            if(fd >= 0)
            {
                int stereo = cfg.channels-1;
                ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &cfg.resolution);
                ioctl(fd, SNDCTL_DSP_STEREO,     &stereo);
                ioctl(fd, SNDCTL_DSP_SPEED,      &cfg.sampling_rate);
            }
        }
        
        if(fd < 0)
        {
            string s = snd_dev + " could not be opened";
            S9xMessage(0,0, s.c_str());
            break;
        }
        
        //memset(Buf, 0, bufsize);
        //write(fd, Buf, bufsize);
        
        fprintf(stderr, "File: %s\n", fn);
        
        SPC_ID666 *id = SPC_get_id666(fn);
        if(id)
        {
            char Unkn[64];sprintf(Unkn,"Unknown (%d)", (int)id->emulator);
            fprintf(stderr,
                 "Song: %-32s     Game: %-32s\n"
                 "Comments: %-32s Author: %-32s\n"
                 "Dumper: %-16s   Playtime: %d   Fadetime: %d   Emulator: %s\n",
                 id->songname,
                 id->gametitle,
                 id->comments,
                 id->author,
                 id->dumper,
                 id->playtime,
                 id->fadetime,
                 (id->emulator==1) ? "ZSnes"
               : (id->emulator==2) ? "snes9x"
               : Unkn);
            free(id);
        }
        else
            fprintf(stderr, "(no id666 tag)\n");
        
        State();
        for(;;)
        {
            unsigned char c;
            if(read(0, &c, 1) == 1)
            {
                if(c=='q')break;
                if(c=='u')surround ^= 1;
                if(c=='a')Settings.AltSampleDecode ^= TRUE;
                if(c=='9' || c=='i')Settings.InterpolatedSound ^= TRUE;
                if(c=='e')Settings.DisableSoundEcho ^= TRUE;
                if(c=='c')if(cur_rate<50)++cur_rate;else cur_rate+=10;
                if(c=='d')if(cur_rate>50)cur_rate-=10;else
                          if(cur_rate>1)--cur_rate;
                if(c=='+')
                {
                    mastervolume *= 1.259921049894873;
                }
                if(c=='-')
                {
                    mastervolume /= 1.259921049894873;
                    if(mastervolume < 0.004)mastervolume = 0.004;
                }
                if(c>='1' && c<='8')
                    S9xSetSoundControl(so.sound_switch ^= (1 << (c-'1')));
                else if(c=='0')
                    S9xSetSoundControl(so.sound_switch = 255);
                State();
            }
            SPC_update(Buf);
            
            /* LITTLE-ENDIAN! */            
            signed short *smp = (signed short *)Buf;
            unsigned count = bufsize / sizeof(*smp);
            for(unsigned a=0; a<count; ++a)
            {
                int v = static_cast<int> (*smp * mastervolume);
                if(surround && (a&1))v=-v;
                if(v > 32767)v = 32767;else if(v < -32768)v = -32768;
                *smp++ = v;
            }
            write(fd, Buf, bufsize);
        }
        close(fd);
    }
    
    free(Buf);
    SPC_close();
    S9xExit();
}
